from collections import namedtuple

CACHE_TTL_ON_UPDATE_SECS = "cache_ttl_on_update_secs"
CACHE_MAX_MEMORY_MB = "cache_max_memory_mb"
CACHE_QUERY_TTL_SECS = "cache_query_ttl_secs"
CACHE_FEEDBACK_TTL_SECS = "cache_feedback_ttl_secs"
ADHOC_EXECUTE = 'adhoc_execute'

DISPLAY_NAME_MAPPING = {
    "Cache TTL on update (sec.)": CACHE_TTL_ON_UPDATE_SECS,
    "Cache max. memory (MB)": CACHE_MAX_MEMORY_MB,
    "Cache Query TTL (sec.)": CACHE_QUERY_TTL_SECS,
    "Cache Feedback TTL (sec.)": CACHE_FEEDBACK_TTL_SECS,
    "Adhoc Execute": ADHOC_EXECUTE
}

def _get_supported_config(schema_version):
    '''
    Get all supported configurations given a schema version
    '''
    if schema_version >= 7:
        # adhoc_execute is added in schema version 7
        return [CACHE_TTL_ON_UPDATE_SECS, CACHE_MAX_MEMORY_MB,
                CACHE_QUERY_TTL_SECS, CACHE_FEEDBACK_TTL_SECS,
                ADHOC_EXECUTE]
    if schema_version == 6:
        # cache_query_ttl_secs and cache_feedback_tl_secs are
        # added in schema version 6
        return [CACHE_TTL_ON_UPDATE_SECS, CACHE_MAX_MEMORY_MB,
                CACHE_QUERY_TTL_SECS, CACHE_FEEDBACK_TTL_SECS]
    else:
        return [CACHE_TTL_ON_UPDATE_SECS, CACHE_MAX_MEMORY_MB]

def _display_name_to_parameter_name(name):
    return DISPLAY_NAME_MAPPING[name]

def _invert_display_name_mapping():
    return {v: k for k, v in DISPLAY_NAME_MAPPING.items()}

def _param_name_to_display_name(parameter_name, mapping=None):
    if not mapping:
        mapping = _invert_display_name_mapping()

    return mapping[parameter_name]

def _config_items_to_kwargs(config_items):
    return {_display_name_to_parameter_name(k): v for k, v in config_items}

class SystemConfig(namedtuple("SystemConfig",
                              [CACHE_TTL_ON_UPDATE_SECS,
                               CACHE_MAX_MEMORY_MB,
                               CACHE_QUERY_TTL_SECS,
                               CACHE_FEEDBACK_TTL_SECS,
                               ADHOC_EXECUTE])):
    def __new__(cls, cache_ttl_on_update_secs=None, cache_max_memory_mb=None,
                cache_query_ttl_secs=None, cache_feedback_ttl_secs=None,
                adhoc_execute=None):
        if cache_ttl_on_update_secs is not None:
            cache_ttl_on_update_secs = int(cache_ttl_on_update_secs)

        if cache_max_memory_mb is not None:
            cache_max_memory_mb = int(cache_max_memory_mb)

        if cache_query_ttl_secs is not None:
            cache_query_ttl_secs = int(cache_query_ttl_secs)

        if cache_feedback_ttl_secs is not None:
            cache_feedback_ttl_secs = int(cache_feedback_ttl_secs)

        return super(SystemConfig, cls).__new__(
            cls, cache_ttl_on_update_secs,
            cache_max_memory_mb,
            cache_query_ttl_secs,
            cache_feedback_ttl_secs,
            adhoc_execute)

    @classmethod
    def from_config_items(cls, config_items):
        """
        Takes the output of a RawConfigParser.items call.
        """
        return SystemConfig(**_config_items_to_kwargs(config_items))

    @classmethod
    def from_config_parser(cls, config_parser, section_name):
        items = config_parser.items(section_name) \
            if config_parser.has_section(section_name) else []
        return cls.from_config_items(items)

    def set_state_in_config(self, config, section_name, schema_version):
        param_name_mapping = _invert_display_name_mapping()
        supported_fields = _get_supported_config(schema_version)
        for field in self._fields:
            if field in supported_fields:
                config.set(section_name,
                           _param_name_to_display_name(field, param_name_mapping),
                           getattr(self, field))

    def for_json(self):
        return {k: v for k, v in self._asdict().items() if v is not None}

    def copy_and_update(self, **kwargs):
        # A tuple is immutable.
        # _replace() returns a new tuple with your modifications.
        return self._replace(**kwargs)
